"""
Module with models for calculating melt Fe\ :sup:`3+`\ /Fe\ :sup:`2+` ratios.
"""

from MagmaPandas.Fe_redox.Fe_redox import FeRedox_QFM
from MagmaPandas.Fe_redox.models import (
    Fe3Fe2_borisov,
    Fe3Fe2_kressCarmichael,
    Fe3Fe2_models,
)



import numpy as np
import scipy.optimize as opt


class EOSparams:
    """
    h                       enthalpy of formation
    s                       entropy
    v0                      volume at 1bar, 298K
    n                       atoms per formula unit
    a0                      coefficient of thermal expansion
    K0                       bulk modulus at 1bar, 298K
    dKdP                    first derivative of K
    dKdP2                   second derivative of K
    cp_a, cp_b, cp_c, cp_d  coefficients of heat capacity polynomial: a + bT + cT**-2 + dT**-(1/2)
    Tc0                     Landau critical temperature at 1 bar
    vmax                    maximum volume of disorder
    smax                    maximum entropy of disorder

    data from Holland & Powell (2011)
    """

    fayalite = {
        "h": -1477.510,
        "s": 151.0,
        "v0": 4.631,
        "n": 7,
        "a0": 2.82e-5,
        "K0": 1256,
        "dKdP": 4.68,
        "dKdP2": -3.7e-3,
        "cp_a": 2.011e-1,
        "cp_b": 1.733e-5,
        "cp_c": -1960.6,
        "cp_d": -9.009e-1,
    }

    ringwoodite = {
        "h": -1477.510,
        "s": 140.0,
        "v0": 4.203,
        "n": 7,
        "a0": 2.22e-5,
        "K0": 1977,
        "dKdP": 4.92,
        "dKdP2": -2.5e-3,
        "cp_a": 1.668e-1,
        "cp_b": 4.2610e-5,
        "cp_c": -1705.4,
        "cp_d": -5.414e-1,
    }

    quartz = {
        "h": -910.710,
        "s": 41.43,
        "v0": 2.269,
        "n": 3,
        "a0": 0,
        "K0": 730,
        "dKdP": 6,
        "dKdP2": -8.2e-3,
        "smax": 4.95 / 1e3,
        "vmax": 1.188e-1,
        "Tc0": 847,
        "cp_a": 9.29e-2,
        "cp_b": -6.42e-7,
        "cp_c": -714.9,
        "cp_d": -0.7161,
    }

    coesite = {
        "h": -906.990,
        "s": 39.60,
        "v0": 2.064,
        "n": 3,
        "a0": 1.23e-5,
        "K0": 979,
        "dKdP": 4.19,
        "dKdP2": -4.3e-3,
        "cp_a": 1.078e-1,
        "cp_b": -3.279e-6,
        "cp_c": -190.3,
        "cp_d": -1.0416,
    }

    stishovite = {
        "h": -876.720,
        "s": 24.0,
        "v0": 1.401,
        "n": 3,
        "a0": 1.58e-5,
        "K0": 3090,
        "dKdP": 4.6,
        "dKdP2": -1.50e-3,
        "cp_a": 6.81e-2,
        "cp_b": 6.010e-6,
        "cp_c": -1978.2,
        "cp_d": -8.21e-2,
    }

    magnetite = {
        "h": -1114.510,
        "s": 146.9,
        "v0": 4.452,
        "n": 7,
        "a0": 3.71e-5,
        "K0": 1857,
        "dKdP": 4.05,
        "dKdP2": -2.2e-3,
        "smax": 35.0,
        "vmax": 0.0,
        "Tc0": 848,
    }


def tait_eos_pressure(phase, pkbar, t, tref=298.15, **kwargs):
    """
    Pressure contribution to Gibb's free energy.
    Tait equation of state from Holland and Powell (2011)

    Parameters
    ----------
    phase       str
        Mineral phase
    pkbar       int, float
        Pressure in kilobar
    t           int, float
        Temperature in Kelvin
    tref        int, float
        Reference temperature in Kelvin

    Returns
    Pth         float
        Thermal pressure term
    a, b, c     float
        Equation of state parameters
    VdP         float
        Contribution of pressure to the Gibbs free energy
    """

    params = ["s", "v0", "n", "a0", "K0", "dKdP", "dKdP2"]
    s, v0, n, a0, K0, dKdP, dKdP2 = [getattr(EOSparams, phase)[i] for i in params]

    # Einstein temperature
    theta = 10636.0 / (s / n + 6.44)

    u0 = theta / tref
    u = theta / t

    # Einstein function page 345
    xi0 = u0**2 * np.exp(u0) / (np.exp(u0) - 1) ** 2.0

    # Equation 3
    a = (1.0 + dKdP) / (1.0 + dKdP + K0 * dKdP2)
    b = dKdP / K0 - dKdP2 / (1.0 + dKdP)
    c = (1.0 + dKdP + K0 * dKdP2) / (dKdP**2.0 + dKdP - K0 * dKdP2)

    # thermal pressure term, equation 11
    Pth = a0 * K0 * theta / xi0 * (1 / (np.exp(u) - 1.0) - 1 / (np.exp(u0) - 1.0))

    # Intergral of volume, equation 13
    PV0 = pkbar * v0
    part1 = np.sign(1 - b * Pth) * abs(1 - b * Pth) ** (1 - c)
    part2 = np.sign(1 + b * (pkbar - Pth)) * abs(1 + b * (pkbar - Pth)) ** (1 - c)
    part3 = b * (c - 1) * pkbar

    VdP = PV0 * (1 - a + a * (part1 - part2) / part3)

    return VdP


def enthalpy(phase, t, tref=298.15):
    """
    Enthalpy as Cp.dT, integrated from T to T(reference), with heat capacity as:

    Cp = a + bT + cT**-2 + dT**(-1/2)

    Cp, a, b, c, & d from Holland and Powell (2011)

    Parameters
    ----------
    phase   str

    t       int, float
        temperature in Kelvin

    tref    int, float
        reference temperature in Kelvin

    Returns
    -------
    enthalpy   float
    """

    a, b, c, d = [
        getattr(EOSparams, phase)[i] for i in ["cp_a", "cp_b", "cp_c", "cp_d"]
    ]

    integral = lambda T: a * T + 0.5 * b * T**2 - c * T**-1 + 2 * d * T ** (1 / 2)
    enthalpy = integral(t) - integral(tref)

    return enthalpy


def entropy(phase, t, tref=298.15):
    """
    Entropy as (Cp/T)dT, integrated from T to T(reference), with:

    Cp/T = a/T + b + cT**-3 + dT**(-3/2)

    Cp, a, b, c & d from Holland and Powell (2011)

    Parameters
    ----------
    phase   str

    t       int, float
        temperature in Kelvin

    tref    int, float
        reference temperature in Kelvin

    Returns
    -------
    entropy   float
    """

    a, b, c, d = [
        getattr(EOSparams, phase)[i] for i in ["cp_a", "cp_b", "cp_c", "cp_d"]
    ]

    integral = lambda T: a * np.log(T) + b * T - c / 2 * T**-2 - 2 * d * T ** (-1 / 2)
    entropy = integral(t) - integral(tref)

    return entropy


def landau(phase, pkbar, T_K, **kwargs):
    """
    Excess Gibbs free energy from Landau theory.

    Based on:
    Holland and Powell, 1998, p.312
    and
    Holland and Powell, 1990

    Rewritten according to Michael Anenberg:
    https://fo2.rses.anu.edu.au/fo2app/
    (see FMQ buffer details and references)

    Parameters
    ----------
    phase   str
        Mineral phase
    pkbar   int, float
        Pressure in kilobar
    t       int, float
        Temperature in Kelvin
    vmax    int, float
        Optional, maximum volume of disorder. Set to 0 to calculate the pressure independent contribution.

    Returns
    -------
    float
        Pressure dependent contribution to the excess Gibbs free enery from Landau theory
    """
    t = np.array([])
    t = np.append(t, T_K)

    vmax_default = getattr(EOSparams, phase)["vmax"]
    vmax = kwargs.get("vmax", vmax_default)

    smax, tc0 = [getattr(EOSparams, phase)[i] for i in ["smax", "Tc0"]]

    Q2_0 = np.sqrt(1 - 298.15 / tc0)

    # Landau critical temperature at pkbar
    tc = tc0 + pkbar * vmax / smax

    Q2 = np.zeros(
        shape=[
            len(t),
        ]
    )
    if any(t < tc):
        Q2 = np.where(t > tc, 0, np.sqrt((tc - t) / tc0))

    G_Landau = (
        smax * (tc0 * (Q2_0 + (Q2**3 - Q2_0**3) / 3) - tc * Q2 - t * (Q2_0 - Q2))
        + pkbar * vmax * Q2_0
    )

    # Convert to float if there is only one item
    if len(G_Landau) == 1:
        G_Landau = G_Landau.item()

    return G_Landau


def landau_P_dependent(phase, pkbar, t, formulation="anenberg"):
    """
    Pressure dependent excess Gibbs free energy from Landau theory.
    Calculated by subtracting the pressure independent contribution (with vmax = 0) from the total.

    Parameters
    ----------
    phase       :   str
        Mineral phase
    t           :   int, float
        Temperature in Kelvin
    pkbar       :   int, float
        Pressure in kilobars
    formulation :   str
        'holland' for the formulation by Holland and Powell (1998), otherwise Michael Anenberg's
        formulation will be used:
        https://fo2.rses.anu.edu.au/fo2app/
        (see FMQ buffer details and references)
    Returns
    -------
    float
        Pressure contribution to excess Gibbs free enery from Landau theory
    """
    if formulation == "holland":
        landau_total = landau_Holland(phase, pkbar, t)
        landau_1bar = landau_Holland(phase, 0, t, vmax=0)
    else:
        landau_total = landau(phase, pkbar, t)
        landau_1bar = landau(phase, 0, t, vmax=0)

    return landau_total - landau_1bar


def landau_Holland(phase, pkbar, T_K, **kwargs):
    """
    Excess Gibbs free energy from Landau theory

    Equations from Holland and Powell (1998), p. 312

    Parameters
    ----------
    phase   str
        Mineral phase
    pkbar   int, float
        Pressure in kilobars
    t       int, float
        Temperature in Kelvin

    Returns
    -------
    float
        Excess Gibbs free enery from Landau theory
    """
    t = np.array([])
    t = np.append(t, T_K)

    vmax_default = getattr(EOSparams, phase)["vmax"]
    vmax = kwargs.get("vmax", vmax_default)
    smax, tc0, a0, K0 = [
        getattr(EOSparams, phase)[i] for i in ["smax", "Tc0", "a0", "K0"]
    ]

    # Landau critical temperature
    tc = tc0 + vmax * pkbar / smax
    # Q: oder paramter in the landau model
    Q2_0 = np.sqrt(1 - 298.15 / tc0)

    # if t > tc:
    #     Q2 = 0
    # else:
    #     Q2 = np.sqrt((tc - t) / tc0)

    Q2 = np.where(t > tc, 0, np.sqrt((tc - t) / tc0))

    # Bulk modulus
    K = K0 * (1 - 1.5e-4 * (t - 298))

    # Excess enthalpy at 298K from Landau model disordering
    h = smax * tc0 * (Q2_0 - (Q2_0**3) / 3)
    # Excess entropy at 298K from Landau model disordering
    s = smax * Q2_0

    # Excess volume from Landau model disordering
    vt = vmax * Q2_0 * (1 + a0 * (t - 298)) - 20 * a0 * (np.sqrt(t) - np.sqrt(298))

    vtdP = vt * K / 3 * ((1 + 4 * pkbar / K) ** (3 / 4) - 1)

    delta_G_landau = smax * ((t - tc0) * Q2 + (tc * Q2**3) / 3)

    G_excess = h - t * s + vtdP + delta_G_landau

    # Convert to float if there is only one item
    if len(G_excess) == 1:
        G_excess = G_excess.item()

    return G_excess


def phaseTransition(pkbar, t, phase_1, phase_2):
    """
    Gibbs free energy of transition from phase_1 to phase_2

    Parameters
    ----------
    pkbar               int, float
        Pressure in kilobar
    t                   int, float
        Temperature in Kelvin
    phase_1, phase_2    str
        Mineral phases

    Returns
    -------
    float
        Gibbs free energy of phase transition
    """

    results = []

    for phase in [phase_1, phase_2]:

        h = getattr(EOSparams, phase)["h"]
        s = getattr(EOSparams, phase)["s"] / 1e3

        Gibbs = h + enthalpy(phase=phase, t=t) - t * (s + entropy(phase=phase, t=t))
        VdP = tait_eos_pressure(phase=phase, pkbar=pkbar, t=t)

        Gibbs = Gibbs + VdP

        if phase in ["quartz", "magnetite"]:

            Gibbs = Gibbs + landau(phase=phase, pkbar=pkbar, T_K=t)

        results.append(Gibbs)

    return results[0] - results[1]



import warnings as w

import numpy as np
import pandas as pd

from MagmaPandas.Fe_redox.Fe_redox_baseclass import Fe3Fe2_model


class Fe3Fe2_borisov(Fe3Fe2_model):
    """
    Calculate melt |Fe3Fe2| ratios according to equation 4 from Borisov et al. (2018)\ [1]_.
    """

    log_Fe3Fe2_error = 0.079

    @classmethod
    def calculate_Fe3Fe2(
        cls, Melt_mol_fractions: pd.DataFrame, T_K, fO2, *args, **kwargs
    ):
        """
        Calculate melt |Fe3Fe2| ratios.

        Parameters
        ----------
        mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
            Melt composition in oxide mol fractions
        T_K :   float, array-like
            temperature in Kelvin
        fO2 :   float, array-like
            Oxygen fugacity

        Returns
        -------
        float, array-like
            melt |Fe3Fe2| ratio
        """

        moles = Melt_mol_fractions.copy()

        oxides = ["SiO2", "TiO2", "MgO", "CaO", "Na2O", "K2O", "Al2O3", "P2O5"]

        if isinstance(moles, pd.DataFrame):
            missing_oxides = set(oxides).difference(moles.columns)
        elif isinstance(moles, pd.Series):
            missing_oxides = set(oxides).difference(moles.index)

        if len(missing_oxides) > 0:

            for oxide in missing_oxides:
                moles[oxide] = 0.0

            w.warn(
                f"{', '.join(str(i) for i in missing_oxides)} missing in composition and set to 0."
            )

        part1 = (
            0.207 * np.log10(fO2)
            + 4633.3 / T_K
            - 0.445 * moles["SiO2"]
            - 0.900 * moles["TiO2"]
            + 1.532 * moles["MgO"]
        )
        part2 = (
            0.314 * moles["CaO"]
            + 2.030 * moles["Na2O"]
            + 3.355 * moles["K2O"]
            - 4.851 * moles["P2O5"]
        )
        part3 = (
            -3.081 * moles["SiO2"] * moles["Al2O3"]
            - 4.370 * moles["SiO2"] * moles["MgO"]
            - 1.852
        )

        return 10 ** (part1 + part2 + part3)

    @classmethod
    def get_error(cls, Fe3Fe2, *args, **kwargs):
        """
        Calculate one standard deviation errors on |Fe3Fe2| ratios, based on a reported 0.079 calibration error on log(|Fe3Fe2|)

        Parameters
        ----------
        Fe3Fe2  :   float, array-like
            melt |Fe3Fe2| ratios

        Returns
        -------
        float, array-like
            |Fe3Fe2| error
        """

        return cls.log_Fe3Fe2_error * Fe3Fe2

    @classmethod
    def get_offset_parameters(cls, n: int = 1):
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, Fe3Fe2, offset_parameters, *args, **kwargs):
        return offset_parameters * cls.log_Fe3Fe2_error * Fe3Fe2



from MagmaPandas.Fe_redox.models.borisov import Fe3Fe2_borisov
from MagmaPandas.Fe_redox.models.kressCarmichael import Fe3Fe2_kressCarmichael

Fe3Fe2_models = {"borisov": Fe3Fe2_borisov, "kressCarmichael": Fe3Fe2_kressCarmichael}



import pandas as pd
from typing_extensions import Self

from MagmaPandas.enums import Unit
from MagmaPandas.MagmaFrames.magmaFrame import MagmaFrame


class Olivine(MagmaFrame):
    """
    Subclass of :py:class:`~MagmaPandas.MagmaFrames.magmaFrame.MagmaFrame` extended with olivine specific methods.
    """

    # @property
    # def _constructor(self):
    #     """This is the key to letting Pandas know how to keep
    #     derivatives of `MagmaBase` the same type as yours.  It should
    #     be enough to return the name of the Class.  However, in
    #     some cases, `__finalize__` is not called and `new attributes` are
    #     not carried over.  We can fix that by constructing a callable
    #     that makes sure to call `__finalize__` every time."""

    #     def _c(*args, weights=None, **kwargs):
    #         if weights is None:
    #             weights = self._weights.copy(deep=True)
    #         return Olivine(*args, weights=weights, **kwargs).__finalize__(self)

    #     return _c

    @property
    def forsterite(self) -> pd.Series:
        """
        Forsterite contents
        """
        if self._units == Unit.WT_PERCENT:
            moles = self.moles
        else:
            moles = self

        type = self._datatype.value

        Mg = {"oxide": "MgO", "cation": "Mg"}
        Fe = {"oxide": "FeO", "cation": "Fe"}
        self.recalculate(inplace=True)
        return pd.Series(
            moles[Mg[type]] / (moles[Fe[type]] + moles[Mg[type]]),
            name="Fo#",
        )

    @property
    def formula(self) -> Self:
        """
        Mineral formulas normalised to 4 O p.f.u.
        """
        return self.mineral_formula(O=4)



"""
Global configuration of MagmaPandas settings.
"""

from MagmaPandas.parse_io.validate import _check_setter

Fe3Fe2_models = ["borisov", "kressCarmichael"]
Kd_ol_FeMg_models = ["toplis", "blundy"]
melt_thermometers = ["putirka2008_14", "putirka2008_15", "putirka2008_16"]
volatile_solubility_models = ["IaconoMarziano", "Allison2022", "Shiskina"]
volatile_species_options = ["co2", "h2o", "mixed"]


_variables = {
    "fO2 buffers": ["QFM"],
    "Melt Fe3+/Fe2+ models": Fe3Fe2_models,
    "Ol-melt Fe-Mg Kd models": Kd_ol_FeMg_models,
    "Melt thermometers": melt_thermometers,
    "Volatile solubility models": volatile_solubility_models,
    "Volatile species": volatile_species_options,
}

_names_length = max([len(i) for i in _variables.keys()]) + 3
_pad_right = max([len(", ".join(i)) for i in _variables.values()]) + 3
_pad_total = _names_length + _pad_right
_new_line = "\n"

configuration_options = (
    f"{_new_line}{' MagmaPandas ':#^{_pad_total}}"
    f"{_new_line}{'':#^{_pad_total}}"
    f"{_new_line}{'Configuration options':_<{_pad_total}}"
)

for param, val in _variables.items():
    configuration_options += (
        f"{_new_line}{param:.<{_names_length}}{', '.join(val):.>{_pad_right}}"
    )


class _meta_configuration(type):
    def __init__(cls, *args, **kwargs):
        cls._dQFM = 1
        cls._Kd_model = "toplis"
        cls._Fe3Fe2_model = "borisov"
        cls._melt_thermometer = "putirka2008_15"
        cls._volatile_solubility = "IaconoMarziano"
        cls._volatile_species = "mixed"

    @property
    def dQFM(cls):
        return cls._dQFM

    @dQFM.setter
    def dQFM(cls, value):
        cls._dQFM = value

    @property
    def Kd_model(cls):
        return cls._Kd_model

    @Kd_model.setter
    @_check_setter(Kd_ol_FeMg_models)
    def Kd_model(cls, model: str):
        cls._Kd_model = model

    @property
    def Fe3Fe2_model(cls):
        return cls._Fe3Fe2_model

    @Fe3Fe2_model.setter
    @_check_setter(Fe3Fe2_models)
    def Fe3Fe2_model(cls, model: str):
        cls._Fe3Fe2_model = model

    @property
    def melt_thermometer(cls):
        return cls._melt_thermometer

    @melt_thermometer.setter
    @_check_setter(melt_thermometers)
    def melt_thermometer(cls, model: str):
        cls._melt_thermometer = model

    @property
    def volatile_solubility(cls):
        return cls._volatile_solubility

    @volatile_solubility.setter
    @_check_setter(volatile_solubility_models)
    def volatile_solubility(cls, model: str):
        cls._volatile_solubility = model

    @property
    def volatile_species(cls):
        return cls._volatile_species

    @volatile_species.setter
    @_check_setter(volatile_species_options)
    def volatile_species(cls, model: str):
        cls._volatile_species = model

    def __str__(cls):

        variables = {
            "\u0394 buffer": "_dQFM",
            "Melt Fe3+/Fe2+": "_Fe3Fe2_model",
            "Kd Fe-Mg ol-melt": "_Kd_model",
            "Melt thermometer": "_melt_thermometer",
            "Volatile solubility model": "_volatile_solubility",
            "Volatile species": "_volatile_species",
        }

        names_length = max([len(i) for i in variables.keys()]) + 5
        pad_right = 15
        pad_total = names_length + pad_right
        new_line = "\n"

        message = (
            f"{new_line}{' MagmaPandas ':#^{pad_total}}"
            f"{new_line}{'':#^{pad_total}}"
            f"{new_line}{'General settings':_<{pad_total}}"
            f"{new_line}{'fO2 buffer':.<{names_length}}{'QFM':.>{pad_right}}"
        )

        parameter_settings = ""
        for param, value in variables.items():
            parameter_settings += (
                f"{new_line}{param:.<{names_length}}{getattr(cls, value):.>{pad_right}}"
            )

        return message + parameter_settings


class configuration(metaclass=_meta_configuration):
    """
    Class for configuring global settings in MagmaPandas.

    Attributes
    ----------
    dQFM    : int, float
        Log units shift of the QFM |fO2| buffer. Default value: 1
    Kd_model : str
        Olivine-melt Fe-Mg partitioning model. Available models: 'toplis'\ [10]_ and 'blundy'\ [11]_. Default value: 'toplis'
    Fe3Fe2_model : str
        Melt |Fe3Fe2| model. Available models: 'borisov'\ [1]_ and 'kressCarmichael'\ [2]_. Default value: 'borisov'
    melt_thermometer : str
        Melt-only thermometer. Available models: 'putirka2008_14'\ [15]_, 'putirka2008_15'\ [15]_ and 'putirka2008_16'\ [15]_. Default value: 'putirka2008_15'
    volatily_solubility : str
        |CO2|-|H2O| solubility model. Available models: 'IaconoMarziano'\ [18]_, 'Allison2022'\ [17]_, 'Shiskina'\ [19]_. Default value: 'IaconoMarziano'
    volatile_species : str
        Fluid phase species. Options: 'h2o', 'co2' or 'mixed'. Default value: 'mixed'
    """

    @classmethod
    def reset(cls):
        """
        Reset to default values
        """
        cls._dQFM = 1
        cls._Kd_model = "toplis"
        cls._Fe3Fe2_model = "borisov"
        cls._melt_thermometer = "putirka2008_15"
        cls._volatile_solubility = "IaconoMarziano"
        cls._volatile_species = "mixed"



import numpy as np
import pandas as pd
from scipy.constants import R  # J*K-1*mol-1

from MagmaPandas.Kd.Kd_baseclass import Kd_model


class FeMg_Toplis(Kd_model):
    """
    Calculate equilibrium Fe-Mg partition coefficients between olivine and melt according to equation 10 from Toplis (2005)\ [10]_.
    """

    error = 0.02

    @classmethod
    def _Phi(cls, molar_SiO2, molar_Na2O, molar_K2O):
        """
        Equation 12 from Toplis (2005) calculates a Phi parameter to correct SiO2 for alkali-bearing liquids.
        This expression is only valid for SiO2 <= 60 mol %


        Parameters
        ----------
        molar_SiO2 : int or list-like

        molar_Na2O : int or list-like

        molar_K2O : int or list-like


        Returns
        -------
            int or list-like
        """

        try:
            if sum(np.array(molar_SiO2) > 60) > 1:
                raise RuntimeError("SiO2 >60 mol% present")
        except:
            if molar_SiO2 > 60:
                raise RuntimeError("SiO2 >60 mol%")

        return (0.46 * (100 / (100 - molar_SiO2)) - 0.93) * (molar_Na2O + molar_K2O) + (
            -5.33 * (100 / (100 - molar_SiO2)) + 9.69
        )

    @classmethod
    def _SiO2_A(cls, melt_mol_fractions):
        """returns adjusted SiO2 for Toplis (2005) Fe-Mg Kd calculations

        Equations 11 and 14 calculate adjusted molar SiO2 by correcting for akalis and water


        Parameters
        ----------
        molar_SiO2 : int or list-like

        molar_Na2O : int or list-like

        molar_K2O : int or list-like

        Phi : int or list-like
            coefficient for alkali correction, needs to be calculated according to Toplis (eq 12, 2005)

        H2O : int or list-like, optional
            wt. %


        Returns
        -------
        int or list-like
        """

        # Calculate melt molar concentrations
        # Molar fractions normalised to 1
        melt_mol_fractions = melt_mol_fractions.fillna(0.0)
        molar_concentrations = melt_mol_fractions * 100

        molar_SiO2 = molar_concentrations["SiO2"]
        molar_Na2O = molar_concentrations["Na2O"]
        molar_K2O = molar_concentrations["K2O"]

        Phi = cls._Phi(molar_SiO2, molar_Na2O, molar_K2O)
        # Equation 11
        SiO2_A = molar_SiO2 + Phi * (molar_Na2O + molar_K2O)

        try:
            # For dataframes
            if "H2O" in molar_concentrations.columns:
                SiO2_A = SiO2_A + 0.8 * molar_concentrations["H2O"]  # equation 14
        except:
            # For series
            if "H2O" in molar_concentrations.index:
                SiO2_A = SiO2_A + 0.8 * molar_concentrations["H2O"]  # equation 14

        return SiO2_A

    @classmethod
    def calculate_Kd(
        cls,
        melt_mol_fractions: pd.DataFrame,
        forsterite: float | pd.Series,
        T_K: float | pd.Series,
        P_bar: float | pd.Series,
        *args,
        **kwargs,
    ) -> float | pd.Series:
        """
        Calculate equilibrium Kds for given melt compositions

        Parameters
        ----------
        melt_mol_fractions : pandas Dataframe
            melt compositions in oxide mol fractions
        forsterite : float, array-like
            initial olivine forsterite contents. Forsterite values are iteratively adjusted and initial values are not necessarily in Fe-Mg equilibrium with melts.
        T_K : float, array-like
            temperatures in Kelvin
        P_bar: float, array-like
            pressures in bar

        Returns
        -------
        Kds : array-like
            Fe-Mg partition coefficients
        """

        SiO2_A = cls._SiO2_A(melt_mol_fractions)

        return np.exp(
            (-6766 / (R * T_K) - 7.34 / R)
            + np.log(0.036 * SiO2_A - 0.22)
            + (3000 * (1 - 2 * forsterite) / (R * T_K))
            + (0.035 * (P_bar - 1) / (R * T_K))
        )

    @classmethod
    def get_error(cls) -> float:
        """
        Calculate one standard deviation errors on Kds

        Returns
        -------
        error:  float
            one standard deviation error
        """

        return cls.error

    @classmethod
    def get_offset_parameters(cls, n=1):
        return np.random.normal(loc=0, scale=1, size=n)

    @classmethod
    def get_offset(cls, offset_parameters, *args, **kwargs):
        error = cls.get_error()
        return offset_parameters * error



from abc import ABC, abstractmethod


class Kd_model(ABC):
    @abstractmethod
    def calculate_Kd(
        cls, Melt_mol_fractions, forsterite_initial, T_K, P_bar, *args, **kwargs
    ):
        pass

    @abstractmethod
    def get_error(cls, *args, **kwargs):
        pass

    @abstractmethod
    def get_offset_parameters(cls, n: int, *args, **kwargs):
        pass

    @abstractmethod
    def get_offset(cls, melt_composition, offset_parameters, *args, **kwargs):
        pass



from MagmaPandas import fO2
from MagmaPandas.configuration import configuration
from MagmaPandas.Fe_redox.models import Fe3Fe2_models


def FeRedox_QFM(mol_fractions, T_K, P_bar, logshift, **kwargs):
    """
    Calculate melt |Fe3Fe2| ratios at the QFM *f*\ O2 buffer, with
    models from Borisov et al. (2018)\ [1]_ or Kress and Carmichael (1991)\ [2]_.

    Model choice is set in the global :py:class:`~MagmaPandas.configuration.configuration` class.

    Parameters
    ----------
    mol_fractions   :   :py:class:`Pandas DataFrame <pandas:pandas.DataFrame>`
        Melt composition in oxide mol_fractions
    T_K  :       float, pd.Series-like
        temperature in Kelvin
    P_bar   :     float, pd.Series-like
        Pressure in bars
    logshift  :  int, pd.Series-like
        log units shift of QFM buffer
    model     :  string
        'kressCarmichael' or 'borisov'

    Returns
    -------
    melt Fe\ :sup:`3+`\ /Fe\ :sup:`2+` ratio
    """

    Fe_model_name = kwargs.get("Fe_model", configuration.Fe3Fe2_model)
    Fe_model = Fe3Fe2_models[Fe_model_name]  # getattr(Fe_redox, Fe_model_name)

    fO2_bar = kwargs.get("fO2", fO2.fO2_QFM(logshift, T_K, P_bar))

    return Fe_model.calculate_Fe3Fe2(mol_fractions, T_K, fO2_bar, P_bar)



import pandas as pd

from MagmaPandas.configuration import configuration
from MagmaPandas.Kd.Ol_melt.Models import Kd_models, equilibrium_forsterite


def calculate_Kd(
    Melt_mol_fractions: pd.Series,
    forsterite_initial,
    T_K,
    Fe3Fe2,
    **kwargs,
):
    """

    Parameters
    ----------
    melt_mol_fractions : pd.DataFrame
        melt composition in oxide mol fraction.
    olivine_forsterite
        forsterite fraction in olivine as Mg * 100 / (Mg + Fe)
    T_K :
        Temperature in Kelvin
    P_bar :
        Pressure in bar
    Fe3Fe2 :
        melt Fe3+/Fe2+ ratio
    """
    model_name = kwargs.pop("model", configuration.Kd_model)
    Kd_model = Kd_models[model_name]

    fo_converge_default = 0.001
    fo_converge = kwargs.pop("fo_converge", fo_converge_default)

    for var, name in zip(
        (forsterite_initial, T_K, Fe3Fe2), ("forsterite_initial", "T_K", "Fe3Fe2")
    ):
        if not isinstance(var, (float, int)):
            raise ValueError(f"{name} should be float or int")

    # initialise Kd
    Kd = Kd_model.calculate_Kd(Melt_mol_fractions, forsterite_initial, T_K, **kwargs)

    Fe2Fe_total = 1 / (1 + Fe3Fe2)
    Fe2Mg = (Melt_mol_fractions["FeO"] * Fe2Fe_total) / Melt_mol_fractions["MgO"]
    forsterite_EQ = equilibrium_forsterite(Kd, Fe2Mg)

    # Difference between observed Fo and equilibrium Fo
    forsterite_delta = abs(forsterite_initial - forsterite_EQ) / forsterite_initial

    iterate = forsterite_delta > fo_converge
    # iterate until equilibrium forsterite content doesn't change any more
    while iterate:

        Kd = Kd_model.calculate_Kd(Melt_mol_fractions, forsterite_EQ, T_K, **kwargs)

        forsterite_initial = forsterite_EQ.copy()
        forsterite_EQ = equilibrium_forsterite(Kd, Fe2Mg)
        forsterite_delta = abs(forsterite_initial - forsterite_EQ) / forsterite_initial
        iterate = forsterite_delta > fo_converge

    return Kd
